from abc import ABCMeta
from ConfigParser import NoOptionError, NoSectionError
from functools import partial
import logging

from lxml import etree, objectify
from lxml.builder import ElementMaker
import requests

from .config import production_config, sandbox_config, CONFIG_PATH
from . import namespaces
from .utils import parser_from_schema


logging.basicConfig()
log = logging.getLogger(__name__)

# I wish I didn't have to hard-code this but there's no service to query, see:
# http://developer.ebay.com/DevZone/merchandising/docs/Concepts/SiteIDToGlobalID.html
SITE_ID_TO_GLOBAL_ID = {
    0: 'EBAY-US',
    2: 'EBAY-ENCA',
    3: 'EBAY-GB',
    15: 'EBAY-AU',
    16: 'EBAY-AT',
    23: 'EBAY-FRBE',
    71: 'EBAY-FR',
    77: 'EBAY-DE',
    100: 'EBAY-MOTOR',
    101: 'EBAY-IT',
    123: 'EBAY-NLBE',
    146: 'EBAY-NL',
    186: 'EBAY-ES',
    193: 'EBAY-CH',
    201: 'EBAY-HK',
    203: 'EBAY-IN',
    205: 'EBAY-IE',
    207: 'EBAY-MY',
    210: 'EBAY-FRCA',
    211: 'EBAY-PH',
    212: 'EBAY-PL',
    216: 'EBAY-SG',
}


NSMAP = {
    'soap': namespaces.SOAP_1_2,
    'soapenv': namespaces.SOAP_1_1,
    'ebl': namespaces.EBAY,
    'xsi': namespaces.XSI,
}


def dicttoxml(obj, factory):
    elements = []
    for key, value in obj.items():
        if hasattr(value, 'items'):
            elements.append(dicttoxml(value, factory))
        elif isinstance(value, (list, tuple)):
            elements.append([
                dicttoxml(sub_val, factory)
                for sub_val in value
            ])
        else:
            el = getattr(factory, key)(unicode(value))
            elements.append(el)
    return elements


class APIBase:
    """
    Abstract: make a concrete sub-class (some APIs are provided below)
    """
    __metaclass__ = ABCMeta

    SCHEMA_URL = None

    PRODUCTION_ENDPOINT = None
    SANDBOX_ENDPOINT = None

    SOAP_VERSION = None

    DEFAULT_NAMESPACE = namespaces.EBAY

    def __init__(self, schema_url=None, sandbox=False, **kwargs):
        # eBay API methods are all CamelCase so it should be safe to set
        # lowercase (or all-caps) attributes on self (see __getattr__)

        self.CONF_PREFIX = self._get_conf_prefix()
        self.sandbox = sandbox
        if sandbox:
            self.config = sandbox_config
        else:
            self.config = production_config

        # use passed in schema first, else try config file,
        # else use class default
        if schema_url is not None:
            self._schema = schema_url
        else:
            try:
                self._schema = self.config.get('soap',
                                               '%s_schema' % self.CONF_PREFIX)
            except (NoOptionError, NoSectionError):
                if self.SCHEMA_URL is None:
                    raise NotImplementedError(
                        'You must give a value for SCHEMA_URL on a sub-class,'
                        ' or define <api name>_schema in the conf file'
                    )
                self._schema = self.SCHEMA_URL

        # make a schema-aware parser (for deserializing responses into objects)
        self.parser, self.version = parser_from_schema(self._schema)

        # determine the service endpoint URI
        try:
            self._endpoint = self.config.get('soap',
                                             '%s_api' % self.CONF_PREFIX)
        except (NoOptionError, NoSectionError):
            if sandbox:
                if self.SANDBOX_ENDPOINT is None:
                    raise NotImplementedError(
                        'You must give a value for SANDBOX_ENDPOINT on a sub-'
                        'class, or define <api name>_api in the conf file'
                    )
                self._endpoint = self.SANDBOX_ENDPOINT
            else:
                if self.SANDBOX_ENDPOINT is None:
                    raise NotImplementedError(
                        'You must give a value for PRODUCTION_ENDPOINT on a '
                        'sub-class, or define <api name>_api in the conf file'
                    )
                self._endpoint = self.PRODUCTION_ENDPOINT

        if self.SOAP_VERSION is None:
            raise NotImplementedError(
                'You must give a value for SOAP_VERSION on a sub-class'
            )

        self.client = requests.Session()
        self.client.headers.update({
            'Content-Type': 'application/soap+xml; charset=utf-8',
            'SOAPAction': '',
        })

        log.info('CONFIG_PATH: %s', CONFIG_PATH)
        self.site_id = (
            kwargs.get('site_id') or self.config.get('site', 'site_id')
        )
        self.app_id = (
            kwargs.get('app_id') or self.config.get('keys', 'app_id')
        )
        self.dev_id = (
            kwargs.get('dev_id') or self.config.get('keys', 'dev_id')
        )
        self.cert_id = (
            kwargs.get('cert_id') or self.config.get('keys', 'cert_id')
        )
        self.token = (
            kwargs.get('token') or self.config.get('auth', 'token')
        )

        self._last_response = None

    def __getattr__(self, name):
        """
        make a SOAP request

        Note:
        eBay API methods are all CamelCase, or otherCamelCase
        """
        if name.lower() == name or name.startswith('_'):
            # avoid accidental API calls in ipython..!
            return super(APIBase, self).__getattr(name)
        return partial(self._execute, name=name)

    def _get_conf_prefix(self):
        # assume the class name ends in 'API'
        return self.__class__.__name__[:-3].lower()

    def _nsmap(self):
        nsmap = NSMAP.copy()
        nsmap[None] = self.DEFAULT_NAMESPACE
        return nsmap

    def get_soap_el_factory(self):
        if self.SOAP_VERSION == '1.1':
            return ElementMaker(
                namespace=namespaces.SOAP_1_1,
                nsmap=self._nsmap()
            )
        elif self.SOAP_VERSION == '1.2':
            return ElementMaker(
                namespace=namespaces.SOAP_1_2,
                nsmap=self._nsmap()
            )
        else:
            raise ValueError(
                "Invalid SOAP_VERSION: {}".format(self.SOAP_VERSION)
            )

    def get_msg_el_factory(self):
        return ElementMaker(nsmap=self._nsmap())

    def get_http_headers(self, name):
        return {}

    def _request(self, name, envelope):
        response = self.client.post(
            url=self._endpoint,
            data=etree.tostring(envelope),
            params=self.get_query_params(name),
            headers=self.get_http_headers(name),
        )
        response.raise_for_status()
        self._last_response = response
        return response

    def _execute(self, name, **kwargs):
        """
        this indirection gives us the same call signature as used by
        ebaysdk-python, i.e. `api.execute('GetUser', {})`

        (our __init__ signature and conf are different still at the moment)
        """
        return self.execute(name, kwargs)

    def execute(self, name, params):
        request_name = '{}Request'.format(name)  # can we rely on this?
        factory = self.get_msg_el_factory()
        body_el = getattr(factory, request_name)
        envelope = self.make_envelope([
            body_el(*dicttoxml(params, factory))
        ])

        log.debug(etree.tostring(envelope, pretty_print=True))

        response = self._request(name, envelope)
        return self.objectify_response(name, response)

    def objectify_response(self, name, response):
        soap_response = etree.fromstring(
            response.text.encode(response.encoding)
        )
        response_name = '{}Response'.format(name)  # can we rely on this?
        body_root = soap_response.xpath(
            "//*[local-name() = '%s']" % response_name
        )[0]

        # tostring->fromstring roundtrip is ugly but otherwise the objectified
        # tree is spoiled by namespaces (would have to use getattr everywhere)
        return objectify.fromstring(
            etree.tostring(body_root),
            parser=self.parser
        )

    def get_soap_header(self):
        return None

    def get_query_params(self, name):
        return {}

    def make_envelope(self, body_elements=None):
        """
        body_elements: <list> of etree.Elements or <None>
        """
        soap = self.get_soap_el_factory()
        body_elements = body_elements or []
        body = soap.Body(*body_elements)
        header = self.get_soap_header()
        if header is not None:
            elements = [header, body]
        else:
            elements = [body]
        return soap.Envelope(
            #{
            #    '{%s}encodingStyle' % namespaces.SOAP_1_2: (
            #        'http://www.w3.org/2001/12/soap-encoding')
            #},
            *elements
        )


class TradingAPI(APIBase):
    SCHEMA_URL = 'http://developer.ebay.com/webservices/latest/ebaySvc.xsd'

    PRODUCTION_ENDPOINT = 'https://api.ebay.com/wsapi'
    SANDBOX_ENDPOINT = 'https://api.sandbox.ebay.com/wsapi'

    SOAP_VERSION = '1.1'

    def __getattr__(self, name):
        method = super(TradingAPI, self).__getattr__(name=name)
        # the method call has to include the API version in the body
        # even though it's also provided on the querystring...
        return partial(method, Version=self.version)

    def get_soap_header(self):
        soap = self.get_soap_el_factory()
        payload = self.get_msg_el_factory()
        return soap.Header(
            payload.RequesterCredentials(
                payload.Credentials(
                    payload.AppID(self.app_id),
                    payload.DevId(self.dev_id),
                    payload.AuthCert(self.cert_id),
                ),
                payload.eBayAuthToken(self.token)
            )
        )

    def get_query_params(self, name):
        # for some reason ebay require some fields from the SOAP request to be
        # repeated as querystring args appended to the service url
        return {
            'callname': name,
            'siteid': self.site_id,
            'appid': self.app_id,
            'version': self.version,
            'routing': 'default',
        }


class PlatformNotificationsAPI(TradingAPI):
    """
    The calls to get and set platform notification preferences are actually
    part of the TradingAPI.

    The notifications received at your callback url can be decoded using the
    `objectify_response` method.
    """
    pass


class ShoppingAPI(APIBase):
    SCHEMA_URL = (
        'http://developer.ebay.com/webservices/latest/ShoppingService.xsd'
    )

    PRODUCTION_ENDPOINT = 'http://open.api.ebay.com/shopping'
    SANDBOX_ENDPOINT = 'http://open.api.sandbox.ebay.com/shopping'

    SOAP_VERSION = '1.1'

    def get_query_params(self, name):
        # for some reason ebay require some fields from the SOAP request to be
        # repeated as querystring args appended to the service url
        return {
            'callname': name,
            'siteid': self.site_id,
            'appid': self.app_id,
            'version': self.version,
            'responseencoding': 'SOAP',
            'requestencoding': 'SOAP',
        }


class FindingAPI(APIBase):
    SCHEMA_URL = (
        'http://developer.ebay.com/webservices/Finding/latest/'
        'FindingService.wsdl'
    )

    PRODUCTION_ENDPOINT = (
        'http://svcs.ebay.com/services/search/FindingService/v1'
    )
    SANDBOX_ENDPOINT = (
        'http://svcs.sandbox.ebay.com/services/search/FindingService/v1'
    )

    SOAP_VERSION = '1.2'

    DEFAULT_NAMESPACE = namespaces.EBAY_SEARCH

    def get_http_headers(self, name):
        site_id = int(self.site_id, 10)
        return {
            'X-EBAY-SOA-OPERATION-NAME': name,
            'X-EBAY-SOA-SERVICE-NAME': 'FindingService',  # this one is genius
            'X-EBAY-SOA-SERVICE-VERSION': self.version,
            'X-EBAY-SOA-GLOBAL-ID': SITE_ID_TO_GLOBAL_ID[site_id],
            'X-EBAY-SOA-SECURITY-APPNAME': self.app_id,
            'X-EBAY-SOA-REQUEST-DATA-FORMAT': 'SOAP',
            'X-EBAY-SOA-MESSAGE-PROTOCOL': 'SOAP12',
        }


class BusinessPoliciesAPI(APIBase):
    SCHEMA_URL = (
        'http://developer.ebay.com/webservices/business-policies/'
        'latest/SellerProfilesManagementService.wsdl'
    )

    PRODUCTION_ENDPOINT = (
        'https://svcs.ebay.com/services/selling/v1/'
        'SellerProfilesManagementService'
    )
    SANDBOX_ENDPOINT = (
        'http://svcs.sandbox.ebay.com/services/selling/v1/'
        'SellerProfilesManagementService'
    )

    SOAP_VERSION = '1.2'

    def get_http_headers(self, name):
        site_id = int(self.site_id, 10)
        return {
            'X-EBAY-SOA-OPERATION-NAME': name,
            'X-EBAY-SOA-SERVICE-VERSION': self.version,
            'X-EBAY-SOA-GLOBAL-ID': SITE_ID_TO_GLOBAL_ID[site_id],
            'X-EBAY-SOA-REQUEST-DATA-FORMAT': 'SOAP',
            'X-EBAY-SOA-MESSAGE-PROTOCOL': 'SOAP12',
            'X-EBAY-SOA-SECURITY-TOKEN': self.token,
        }



import re

from appdirs import user_cache_dir
from lxml import etree, objectify
from dogpile.cache import make_region

from . import namespaces
from .version import __version__


region = make_region().configure(
    'dogpile.cache.dbm',
    expiration_time=1209600,  # 14 days
    arguments={
        "filename": "{dir}/{version}.dbm".format(
            dir=user_cache_dir('anentropic', 'dirtyebay'),
            version='dirtyebay_{}'.format(__version__)
        )
    }
)


VERSION_COMMENT = re.compile(r'\s*Version\s*(\d+)\s*')

NS_MAP = {
    'wsdl': namespaces.WSDL,
    'ebay': namespaces.EBAY,
    'xs': namespaces.XSD,
}


class VersionNotFound(Exception):
    pass


def version_from_schema(schema_el):
    """
    returns: API version number <str>
    raises: <VersionNotFound>

    NOTE:
    relies on presence of comment tags in the XSD, which are currently present
    for both ebaySvc.xsd (TradingAPI) and ShoppingService.xsd (ShoppingAPI)
    """
    vc_el = schema_el
    while True:
        vc_el = vc_el.getprevious()
        if vc_el is None:
            break
        if vc_el.tag is etree.Comment:
            match = VERSION_COMMENT.search(vc_el.text)
            if match:
                try:
                    return match.group(1)
                except IndexError:
                    pass
    raise VersionNotFound('Version comment not found preceeding schema node')


def _default_ns_prefix(nsmap):
    """
    XML doc may have several prefix:namespace_url pairs, can also specify
    a namespace_url as default, tags in that namespace don't need a prefix
    NOTE:
    we rely on default namespace also present in prefixed form, I'm not sure if
    this is an XML certainty or a quirk of the eBay WSDLs

    in our case the WSDL contains:
        <wsdl:documentation>
            <Version>1.0.0</Version>
        </wsdl:documentation>
    ...but our query needs to give a prefix to the path of `Version` so we need
    to determine the default namespace of the doc, find the matching prefix and
    return it
    """
    if None in nsmap:
        default_url = nsmap[None]
        prefix = None
        for key, val in nsmap.iteritems():
            if val == default_url and key is not None:
                prefix = key
                break
        else:
            raise ValueError(
                "Default namespace {url} not found as a prefix".format(
                    url=default_url
                )
            )
        return prefix
    raise ValueError("No default namespace found in map")


def version_from_wsdl(wsdl_tree):
    """
    returns: API version number <str>
    raises: <VersionNotFound>

    NOTE:
    relies on presence of documentation node in the WSDLs:
        <wsdl:documentation>
            <Version>1.0.0</Version>
        </wsdl:documentation>
    """
    prefix = _default_ns_prefix(wsdl_tree.nsmap)

    # XPath doesn't allow empty prefix:
    safe_map = wsdl_tree.nsmap.copy()
    try:
        del safe_map[None]
    except KeyError:
        pass

    try:
        # various eBay WSDLs are inconsistent - need case-insensitive matching
        version_el = wsdl_tree.xpath(
            'wsdl:service/wsdl:documentation/'
            '*[self::{p}:version or self::{p}:Version]'.format(p=prefix),
            namespaces=safe_map
        )[0]
    except IndexError:
        raise VersionNotFound(
            'Version not found in WSDL service documentation'
        )
    else:
        return version_el.text


def _make_safe(schema_el):
    """
    Workaround for segfault bug
    https://bugs.launchpad.net/lxml/+bug/1415907

    yes, tostring->fromstring roundtrip is ugly and inefficient
    """
    return etree.fromstring(etree.tostring(schema_el))


@region.cache_on_arguments()
def parser_from_schema(schema_url, require_version=True):
    """
    Returns an XSD-schema-enabled lxml parser from a WSDL or XSD

    `schema_url` can of course be local path via file:// url

    NOTE:
    currently we're not making use of the resulting parser(!) due to:
    https://bugs.launchpad.net/lxml/+bug/1416853
    """
    schema_tree = etree.parse(schema_url)

    def get_version(element, getter):
        try:
            return getter(element)
        except VersionNotFound:
            if require_version:
                raise
            else:
                return None

    root = schema_tree.getroot()
    if root.tag == '{%s}definitions' % namespaces.WSDL:
        # wsdl should contain an embedded schema
        schema_el = schema_tree.find('wsdl:types/xs:schema', namespaces=NS_MAP)
        # hack to avoid segfault:
        schema_el = _make_safe(schema_el)
        version = get_version(root, version_from_wsdl)
    else:
        schema_el = root
        version = get_version(schema_el, version_from_schema)

    schema = etree.XMLSchema(schema_el)
    return objectify.makeparser(schema=schema), version



import os
from ConfigParser import SafeConfigParser


CONFIG_PATH = os.environ.get('DIRTYEBAY_CONFIG_PATH',
                             os.path.abspath('ebay.conf'))
SANDBOX_CONFIG_PATH = os.environ.get('DIRTYEBAY_SANDBOX_CONFIG_PATH',
                                     os.path.abspath('ebay.sandbox.conf'))

production_config = SafeConfigParser()
production_config.read(CONFIG_PATH)

sandbox_config = SafeConfigParser()
sandbox_config.read(SANDBOX_CONFIG_PATH)



EBAY = 'urn:ebay:apis:eBLBaseComponents'
EBAY_SEARCH = 'http://www.ebay.com/marketplace/search/v1/services'
SOAP_1_1 = 'http://schemas.xmlsoap.org/soap/envelope/'
SOAP_1_2 = 'http://www.w3.org/2003/05/soap-envelope'
XSD = 'http://www.w3.org/2001/XMLSchema'
XSI = 'http://www.w3.org/2001/XMLSchema-instance'
WSDL = 'http://schemas.xmlsoap.org/wsdl/'



from .client import (  # noqa
    TradingAPI,
    PlatformNotificationsAPI,
    ShoppingAPI,
    FindingAPI,
    BusinessPoliciesAPI
)
from .config import production_config, sandbox_config  # noqa
from .version import __version__  # noqa



from distutils.core import setup

with open('dirtyebay/version.py') as f:
    # get __version__ var into our scope without importing
    exec(f.read())

setup(
    name='dirtyebay',
    version=__version__,  # noqa
    packages=['dirtyebay'],
    license='MIT',
    long_description=open('pypi.rst').read(),
    description='An eBay API client which respects (XSD) schema and talks SOAP but doesn\'t use Suds',
    author='Anentropic',
    author_email='ego@anentropic.com',
    url='https://github.com/anentropic/dirtyebay',
    install_requires=[
        'lxml',
        'dogpile.cache',
        'appdirs',
        'requests',
    ],
)
